// Implementations of Diffie–Hellman key exchange
use num_bigint::BigUint;
use num_bigint::RandBigInt;
use rand::Rng;
use std::fs;

#[derive(Clone)]
pub struct PublicKey(pub BigUint);

#[derive(Clone)]
pub struct PrivateKey(pub BigUint);

#[derive(Clone)]
pub struct Keypair {
    pub private: PrivateKey,
    pub public: PublicKey,
}

impl Keypair {
    pub fn make(p: &BigUint, g: &BigUint) -> Self {
        let mut rng = rand::thread_rng();
        let private = rng.gen_biguint_below(p);
        let public = g.modpow(&private, p);
        Self {
            private: PrivateKey(private),
            public: PublicKey(public),
        }
    }
}

fn expmod(base: u32, exp: u32, m: u32) -> u32 {
    if exp == 0 {
        return 1;
    }

    if exp % 2 == 0 {
        let s = expmod(base, exp / 2, m);
        (s * s) % m
    } else {
        (expmod(base, exp - 1, m) * base) % m
    }
}

pub fn load_large_prime() -> BigUint {
    let s = fs::read_to_string("data/33.txt").expect("run from git root dir");
    let b: Vec<u8> = s
        .chars()
        .filter(|c| *c != '\n')
        .map(|c| c.to_digit(16).unwrap().try_into().unwrap())
        .collect();
    BigUint::from_radix_be(&b, 16).unwrap()
}

pub fn simple() {
    #![allow(non_snake_case)]
    let mut rng = rand::thread_rng();

    // Set a variable "p" to 37 and "g" to 5.
    let p: u32 = 37;
    let g: u32 = 5;

    // Generate "a", a random number mod 37. Now generate "A", which is "g" raised to the "a"
    // power mode 37 --- A = (g**a) % p.
    let a: u32 = rng.gen::<u32>() % p;
    let A = expmod(g, a, p);

    // Do the same for "b" and "B".
    let b: u32 = rng.gen::<u32>() % p;
    let B = expmod(g, b, p);

    // "A" and "B" are public keys. Generate a session key with them; set "s" to "B" raised to
    // the "a" power mod 37 --- s = (B**a) % p.
    let s = expmod(B, a, p);

    // Do the same with A**b, check that you come up with the same "s".
    let s_ = expmod(A, b, p);

    assert_eq!(s, s_, "crypto is broken");
    // To turn "s" into a key, you can just hash it to create 128 bits of key material (or
    // SHA256 it to create a key for encrypting and a key for a MAC).
}

pub fn serious() {
    let p = load_large_prime();
    let g = 2_u8.into();

    let a = Keypair::make(&p, &g);
    let b = Keypair::make(&p, &g);

    let s = b.public.0.modpow(&a.private.0, &p);
    let s_ = a.public.0.modpow(&b.private.0, &p);
    assert_eq!(s, s_, "crypto is broken");
}
