use crate::{bytes::Bytes, bytes_base64::BytesBase64};
use num_bigint::BigUint;
use openssl::bn::{BigNum, BigNumContext};
use std::{
    io::{BufRead, BufReader},
    time::{SystemTime, UNIX_EPOCH},
};

pub fn unix_timestamp() -> u32 {
    let start = SystemTime::now();
    let since_the_epoch = start
        .duration_since(UNIX_EPOCH)
        .expect("Time went backwards");
    since_the_epoch
        .as_secs()
        .try_into()
        .expect("Linux time no longer fits into 32 bits.")
}

pub fn read_base64(path: &str) -> Bytes {
    let s = std::fs::read_to_string(path).unwrap();
    match BytesBase64::from_base64(&s) {
        Ok(bytes) => bytes.to_bytes(),
        Err(msg) => panic!("Failed to read_base64 {}: {}", path, msg),
    }
}

pub fn read_base64_lines(path: &str) -> Vec<Bytes> {
    let file = std::fs::File::open(path).unwrap();
    let br = BufReader::new(file);
    br.lines()
        .map(|line| {
            BytesBase64::from_base64(
                &line.unwrap_or_else(|_| panic!("Failed to read line in {}", path)),
            )
            .unwrap_or_else(|_| panic!("Invalid base64 in {}", path))
            .to_bytes()
        })
        .collect()
}

pub fn read_hex_lines(path: &str) -> Vec<Bytes> {
    let file = std::fs::File::open(path).unwrap();
    let br = BufReader::new(file);
    br.lines()
        .map(|line| Bytes::from_hex(&line.unwrap()))
        .collect()
}

pub fn xor(a: &[u8], b: &[u8]) -> Vec<u8> {
    Iterator::zip(a.iter(), b.iter())
        .map(|z| *(z.0) ^ *(z.1))
        .collect()
}

pub fn bnclone(b: &BigNum) -> BigNum {
    BigNum::from_slice(&b.to_vec()).unwrap()
}

pub fn cube_root(n: &BigNum) -> BigNum {
    let b = BigUint::from_bytes_be(&n.to_vec());
    let b = b.nth_root(3);
    BigNum::from_slice(&b.to_bytes_be()).unwrap()
}

pub fn bn(n: u32) -> BigNum {
    BigNum::from_u32(n).unwrap()
}

pub fn div_rem(a: &BigNum, b: &BigNum) -> (BigNum, BigNum) {
    // kReturns the divider and remainder of a / b.
    let mut ctx = BigNumContext::new().unwrap();
    let mut div = BigNum::new().unwrap();
    let mut rem = BigNum::new().unwrap();
    div.div_rem(&mut rem, a, b, &mut ctx).unwrap();
    (div, rem)
}

pub fn div_floor(a: &BigNum, b: &BigNum) -> BigNum {
    let (div, _) = div_rem(a, b);
    div
}

pub fn div_ceil(a: &BigNum, b: &BigNum) -> BigNum {
    let (mut div, rem) = div_rem(a, b);
    if rem.num_bits() != 0 {
        div.add_word(1).unwrap();
    }
    div
}
